from bisect import bisect_left, bisect_right
from datetime import datetime
from typing import Optional

from .utils import read_trade_days


class TradeCalendar:
    def __init__(self) -> None:
        self._trade_days: list = read_trade_days()
        self._trade_days_set: set = set(self._trade_days)
        self._trade_days_len: int = len(self._trade_days)

    def is_trade_date(self, date=None) -> bool:
        return _format_date(date) in self._trade_days_set

    def trade_date_list(self) -> list:
        return self._trade_days

    def pre_trade_date(self, date=None) -> Optional[str]:
        date = _format_date(date)
        idx = bisect_left(self._trade_days, date) - 1
        if idx == -1:
            return None
        return self._trade_days[idx]

    def next_trade_date(self, date=None) -> Optional[str]:
        date = _format_date(date)
        idx = bisect_right(self._trade_days, date)
        if idx == len(self._trade_days):
            return None
        return self._trade_days[idx]


def _format_date(date) -> str:
    if date is None:
        return datetime.now().strftime("%Y%m%d")
    if isinstance(date, datetime):
        date = date.strftime("%Y%m%d")
    elif isinstance(date, int):
        date = str(date)
    return date



import abc
import concurrent.futures
from typing import Optional

import requests

from fastquotes.const import HEADERS, REQ_CODES_NUM_MAX
from fastquotes.utils import format_stock_code, format_stock_codes


class Quote(metaclass=abc.ABCMeta):
    def __init__(self):
        self._session = requests.session()

    @property
    @abc.abstractmethod
    def base_url(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def split_char(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def pclose_field_id(self) -> int:
        pass

    @abc.abstractmethod
    def parse_out_tick_dict(self, msg: str) -> Optional[dict]:
        pass

    def current_price(self, code: str) -> float:
        return float(self._detail_list(code)[3])

    def pre_close(self, code: str) -> float:
        return float(self._detail_list(code)[self.pclose_field_id])

    def tick(self, code: str) -> Optional[dict]:
        code = format_stock_code(code)
        tick_dict = self.tick_dict([code])
        tick_list = list(tick_dict.items())
        if not tick_list or not tick_list[0][1]:
            return None
        return tick_list[0][1]

    def price_dict(self, codes: list) -> dict:
        tick_dict = self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "current_price" in tick:
                res_dict[code] = tick["current_price"]
        return res_dict

    def pre_close_dict(self, codes: list) -> dict:
        tick_dict = self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "pre_close" in tick:
                res_dict[code] = tick["pre_close"]
        return res_dict

    def open_dict(self, codes: list) -> dict:
        tick_dict = self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "open" in tick:
                res_dict[code] = tick["open"]
        return res_dict

    def total_vol_dict(self, codes: list) -> dict:
        tick_dict = self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "total_vol" in tick:
                res_dict[code] = tick["total_vol"]
        return res_dict

    def tick_dict(self, codes: list) -> dict:
        format_codes = format_stock_codes(codes)
        res = {}

        def small_price_dict(small_codes: list):
            data_str = self._fetch_codes_data_str(small_codes)
            data_list = data_str.strip().split("\n")
            for item in data_list:
                tick_dict = self.parse_out_tick_dict(item)
                if tick_dict is None:
                    return
                res[tick_dict["code"]] = tick_dict

        codes_len = len(format_codes)
        workers = codes_len // REQ_CODES_NUM_MAX + codes_len % REQ_CODES_NUM_MAX
        with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
            for i in range(0, codes_len, REQ_CODES_NUM_MAX):
                if i + REQ_CODES_NUM_MAX >= codes_len:
                    small_codes = format_codes[i:]
                else:
                    small_codes = format_codes[i : i + REQ_CODES_NUM_MAX]
                executor.submit(small_price_dict, small_codes)
        return res

    def _detail_list(self, code: str) -> list:
        format_code = format_stock_code(code)
        return self._fetch_data_str(format_code).split(self.split_char)

    def _fetch_data_str(self, code: str) -> str:
        response = self._session.get(self.base_url + code, headers=HEADERS)
        return response.text

    def _fetch_codes_data_str(self, codes: list) -> str:
        codes_str = ",".join(codes)
        response = self._session.get(self.base_url + codes_str, headers=HEADERS)
        return response.text



import json

import requests


def fund_intro_dict() -> dict:
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36"
    }
    url = "http://fund.eastmoney.com/js/fundcode_search.js"
    res = requests.get(url, headers=headers)
    text_data = res.text
    res_list = json.loads(text_data.strip("var r = ")[:-1])
    res_dict = {}
    for item in res_list:
        res_dict[item[0]] = {"基金代码": item[0], "基金简称": item[2], "基金类型": item[3]}
    return res_dict


def etf_list() -> list:
    url = (
        "http://vip.stock.finance.sina.com.cn/quotes_service/api"
        "/jsonp.php/IO.XSRV2.CallbackList['da_yPT46_Ll7K6WD']:"
        "/Market_Center.getHQNodeDataSimple"
    )
    params = {
        "page": "1",
        "num": "1000",
        "sort": "symbol",
        "asc": "0",
        "node": "etf_hq_fund",
        "[object HTMLDivElement]": "qvvne",
    }
    r = requests.get(url, params=params)
    data_text = r.text
    data_list = json.loads(data_text[data_text.find("([") + 1 : -2])
    return [item["symbol"] for item in data_list]



from fastquotes.async_quotes.async_sina import AsyncSinaQuote
from fastquotes.async_quotes.async_tencent import AsyncTencentQuote
from fastquotes.fund import etf_list, fund_intro_dict
from fastquotes.fund.daily import fund_real_time_dict
from fastquotes.quotes.sina import SinaQuote
from fastquotes.quotes.tencent import TencentQuote

from .stock_list import (
    async_exchange_stock_list,
    async_stock_list,
    exchange_stock_list,
    stock_list,
    stock_list_sh,
    stock_list_sz,
)
from .to_hans import to_hans, to_hans_amount
from .trade_calendar import TradeCalendar
from .utils import exchange_prefix

__version__ = "0.1.3"
__author__ = "nocilantro"



import os

SINA_BASE_URL = "http://hq.sinajs.cn/list="
TENCENT_BASE_URL = "http://qt.gtimg.cn/q="
REQ_CODES_NUM_MAX = 300
HEADERS = {
    "Accept-Encoding": "gzip, deflate, sdch",
    "User-Agent": (
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/54.0.2840.100 "
        "Safari/537.36"
    ),
}
TRADE_DAYS_FILE_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "static/trade_days.txt")
)



import abc
import asyncio
from typing import Optional

import aiohttp

from fastquotes.const import HEADERS, REQ_CODES_NUM_MAX
from fastquotes.utils import format_stock_codes


class AsyncQuote(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        pass

    @property
    @abc.abstractmethod
    def base_url(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def split_char(self) -> str:
        pass

    @abc.abstractmethod
    def parse_out_tick_dict(self, msg: str) -> Optional[dict]:
        pass

    async def tick_dict(self, codes: list) -> dict:
        format_codes = format_stock_codes(codes)
        res = {}
        tasks = []

        async with aiohttp.ClientSession() as session:

            async def small_price_dict(small_codes: list):
                data_str = await self._fetch_data_str(session, small_codes)
                data_list = data_str.strip().split("\n")
                for item in data_list:
                    tick_dict = self.parse_out_tick_dict(item)
                    if tick_dict is None:
                        return
                    res[tick_dict["code"]] = tick_dict

            codes_len = len(format_codes)
            for i in range(0, codes_len, REQ_CODES_NUM_MAX):
                if i + REQ_CODES_NUM_MAX >= codes_len:
                    small_codes = format_codes[i:]
                else:
                    small_codes = format_codes[i : i + REQ_CODES_NUM_MAX]
                tasks.append(small_price_dict(small_codes))
            await asyncio.wait(tasks)
        return res

    async def price_dict(self, codes: list) -> dict:
        tick_dict = await self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "current_price" in tick:
                res_dict[code] = tick["current_price"]
        return res_dict

    async def pre_close_dict(self, codes: list) -> dict:
        tick_dict = await self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "pre_close" in tick:
                res_dict[code] = tick["pre_close"]
        return res_dict

    async def open_dict(self, codes: list) -> dict:
        tick_dict = await self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "open" in tick:
                res_dict[code] = tick["open"]
        return res_dict

    async def total_vol_dict(self, codes: list) -> dict:
        tick_dict = await self.tick_dict(codes)
        res_dict = {}
        for code, tick in tick_dict.items():
            if "total_vol" in tick:
                res_dict[code] = tick["total_vol"]
        return res_dict

    async def _fetch_data_str(self, session, codes: list) -> str:
        codes_str = ",".join(codes)
        async with await session.get(
            f"{self.base_url}{codes_str}", headers=HEADERS
        ) as response:
            return await response.text()



hans_num_map = ["零", "一", "二", "三", "四", "五", "六", "七", "八", "九"]
hans_units_map = ["千", "百", "十", ""]
amount_hans_num_map = ["零", "壹", "贰", "叁", "肆", "伍", "陆", "柒", "捌", "玖"]
amount_hans_units_map = ["仟", "佰", "拾", ""]
amount_hans_float_units_map = ["角", "分", "厘", "毫"]


def to_hans(s) -> str:
    s = str(s)
    point_pos = s.find(".")
    if point_pos == -1:
        int_part, float_part = s, ""
    else:
        int_part, float_part = s[:point_pos], s[point_pos + 1 :]

    hans = _int_part_hans(int_part, hans_num_map, hans_units_map)

    if float_part:
        hans += "点"
        for num in float_part:
            hans += hans_num_map[int(num)]

    return hans


def to_hans_amount(s) -> str:
    s = str(s)
    point_pos = s.find(".")
    if point_pos == -1:
        int_part, float_part = s, ""
    else:
        int_part, float_part = s[:point_pos], s[point_pos + 1 :]

    hans = _int_part_hans(int_part, amount_hans_num_map, amount_hans_units_map)

    hans += "元"
    float_len = min(len(float_part), 4)
    if float_len == 0:
        hans += "整"
    for i in range(float_len):
        if float_part[i] != "0":
            hans += (
                amount_hans_num_map[int(float_part[i])] + amount_hans_float_units_map[i]
            )

    return hans


def _convert_4_digit(x: str, num_map: list, units_map: list) -> str:
    res = ""
    for i in range(4):
        if x[i] == "0":
            if not res or res[-1] != "零":
                res += "零"
        else:
            res += num_map[int(x[i])] + units_map[i]
    if len(res) > 1 and res[-1] == "零":
        res = res[:-1]
    return res


def _convert_8_digit(x: str, num_map: list, units_map: list) -> str:
    first = _convert_4_digit(x[:4], num_map, units_map)
    second = _convert_4_digit(x[4:], num_map, units_map)
    if first == "零":
        return second
    if second == "零":
        return first + "万"
    return first + "万" + second


def _int_part_hans(int_part: str, num_map: list, units_map: list) -> str:
    remainder = len(int_part) % 8
    if remainder != 0:
        int_part = "0" * (8 - remainder) + int_part

    hans = ""
    split_cnt = len(int_part) // 8
    for i in range(split_cnt):
        hans += _convert_8_digit(
            int_part[i * 8 : i * 8 + 8], num_map, units_map
        ) + "亿" * (split_cnt - i - 1)
    if hans[0] == "零":
        hans = hans[1:]
    return hans



import asyncio
import concurrent.futures
import json

import aiohttp
import requests
import xlrd

from fastquotes.utils import exchange_prefix

url_sh = "http://query.sse.com.cn/security/stock/getStockListData.do"
headers_sh = {
    "Host": "query.sse.com.cn",
    "Pragma": "no-cache",
    "Referer": "http://www.sse.com.cn/assortment/stock/list/share/",
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
        " (KHTML, like Gecko) Chrome/81.0.4044.138 Safari/537.36"
    ),
}
params_sh = {
    "jsonCallBack": "jsonpCallback66942",
    "isPagination": "true",
    "stockCode": "",
    "csrcCode": "",
    "areaName": "",
    "stockType": 1,  # 1表示主板A股,8表示科创板
    "pageHelp.cacheSize": "1",
    "pageHelp.beginPage": "1",
    "pageHelp.pageSize": "2000",
    "pageHelp.pageNo": "1",
    "pageHelp.endPage": "11",
    "_": "1589881387934",
}
url_sz = "http://www.szse.cn/api/report/ShowReport"
params_sz = {
    "SHOWTYPE": "xlsx",
    "CATALOGID": "1110",
    "TABKEY": "tab1",
    "random": "0.6935816432433362",
}


async def async_exchange_stock_list():
    res_list = await async_stock_list()
    return [f"{exchange_prefix(code)}{code}" for code in res_list]


async def async_stock_list():
    res_list = []
    async with aiohttp.ClientSession() as session:

        async def req_sh_list(stock_type):
            params_sh["stockType"] = stock_type
            async with session.get(
                url_sh, params=params_sh, headers=headers_sh
            ) as response:
                text = await response.text()
            loads_data = json.loads(text[text.find("{") : -1])
            res_list.extend([data["SECURITY_CODE_A"] for data in loads_data["result"]])

        async def req_sz_list():

            async with session.get(url_sz, params=params_sz) as r:
                sheet = xlrd.open_workbook(
                    file_contents=(await r.read())
                ).sheet_by_index(0)
                res_list.extend([c.value for c in sheet.col(4)[1:]])

        await asyncio.wait([req_sh_list(1), req_sh_list(8), req_sz_list()])

        return res_list


def exchange_stock_list() -> list:
    res_list = stock_list()
    return [f"{exchange_prefix(code)}{code}" for code in res_list]


def stock_list() -> list:
    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        fut1 = executor.submit(stock_list_sz)
        fut2 = executor.submit(stock_list_sh)
    return fut1.result() + fut2.result()


def stock_list_sz() -> list:
    r = requests.get(url_sz, params=params_sz)
    sheet = xlrd.open_workbook(file_contents=r.content).sheet_by_index(0)
    return [c.value for c in sheet.col(4)[1:]]


def stock_list_sh() -> list:
    def req_list(stock_type):
        params_sh["stockType"] = stock_type
        response = requests.get(url_sh, params=params_sh, headers=headers_sh)
        text = response.text
        loads_data = json.loads(text[text.find("{") : -1])
        res_list = [data["SECURITY_CODE_A"] for data in loads_data["result"]]
        return res_list

    return req_list(1) + req_list(8)



import demjson
import requests


def fund_real_time_dict() -> dict:
    headers = {
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36"
        )
    }
    url = "http://fund.eastmoney.com/Data/Fund_JJJZ_Data.aspx"
    params = {
        "t": "1",
        "lx": "1",
        "letter": "",
        "gsid": "",
        "text": "",
        "sort": "zdf,desc",
        "page": "1,20000",
        "dt": "1580914040623",
        "atfc": "",
        "onlySale": "0",
    }
    res = requests.get(url, params=params, headers=headers)
    text = res.text.strip("var db=")
    data_dict = demjson.decode(text)
    day, pre_day = data_dict["showday"]
    res_dic = {}
    for item in data_dict["datas"]:
        code = item[0]
        res_dic[code] = {
            "code": code,
            "name": item[1],
            "update_date": day,
            "pre_update_date": pre_day,
            "单位净值": item[3],
            "累计净值": item[4],
            "上个交易日单位净值": item[5],
            "上个交易日累计净值": item[6],
            "日涨幅": item[8],
            "手续费": item[17],
            "申购状态": item[9],
            "赎回状态": item[10],
        }
    return res_dic



from typing import Optional

from fastquotes.const import SINA_BASE_URL
from fastquotes.utils import parse_out_sina_tick_dict

from . import quote


class SinaQuote(quote.Quote):
    @property
    def base_url(self) -> str:
        return SINA_BASE_URL

    def parse_out_tick_dict(self, msg: str) -> Optional[dict]:
        return parse_out_sina_tick_dict(msg)

    @property
    def split_char(self) -> str:
        return ","

    @property
    def pclose_field_id(self) -> int:
        return 2



from typing import Optional

from .const import TRADE_DAYS_FILE_PATH


def read_trade_days():
    with open(TRADE_DAYS_FILE_PATH, "r") as f:
        return f.read().split(",")


def parse_out_tencent_tick_dict(msg: str) -> Optional[dict]:
    field_list = msg.split("~")
    try:
        res = {
            "time": field_list[30][8:],
            "name": field_list[1],
            "code": field_list[2],
            "current_price": float(field_list[3]),
            "pre_close": float(field_list[4]),
            "open": float(field_list[5]),
            "high": float(field_list[33]),
            "low": float(field_list[34]),
            "total_amount": float(field_list[37]) * 10000,
            "total_vol": float(field_list[6]) * 100,
            "bid1": float(field_list[9]),
            "bid1_vol": int(field_list[10]) * 100,
            "bid2": float(field_list[11]),
            "bid2_vol": int(field_list[12]) * 100,
            "bid3": float(field_list[13]),
            "bid3_vol": int(field_list[14]) * 100,
            "bid4": float(field_list[15]),
            "bid4_vol": int(field_list[16]) * 100,
            "bid5": float(field_list[17]),
            "bid5_vol": int(field_list[18]) * 100,
            "ask1": float(field_list[19]),
            "ask1_vol": int(field_list[20]) * 100,
            "ask2": float(field_list[21]),
            "ask2_vol": int(field_list[22]) * 100,
            "ask3": float(field_list[23]),
            "ask3_vol": int(field_list[24]) * 100,
            "ask4": float(field_list[25]),
            "ask4_vol": int(field_list[26]) * 100,
            "ask5": float(field_list[27]),
            "ask5_vol": int(field_list[28]) * 100,
            "up_limit": float(field_list[47]),
            "down_limit": float(field_list[48]),
            "price_change": float(field_list[31]),
            "pct_change": float(field_list[32]),
            "flow_market_value": float(field_list[44]),
            "total_market_value": float(field_list[45]),
        }
    except IndexError:
        return None
    return res


def parse_out_sina_tick_dict(msg: str) -> Optional[dict]:
    field_list = msg.split(",")
    try:
        code_name_part = field_list[0].partition('="')
        res = {
            "time": field_list[31].replace(":", ""),
            "name": code_name_part[2],
            "code": code_name_part[0][-6:],
            "current_price": float(field_list[3]),
            "pre_close": float(field_list[2]),
            "open": float(field_list[1]),
            "high": float(field_list[4]),
            "low": float(field_list[5]),
            "total_amount": float(field_list[9]),
            "total_vol": float(field_list[8]),
            "bid1_vol": int(field_list[10]),
            "bid1": float(field_list[11]),
            "bid2_vol": int(field_list[12]),
            "bid2": float(field_list[13]),
            "bid3_vol": int(field_list[14]),
            "bid3": float(field_list[15]),
            "bid4_vol": int(field_list[16]),
            "bid4": float(field_list[17]),
            "bid5_vol": int(field_list[18]),
            "bid5": float(field_list[19]),
            "ask1_vol": int(field_list[20]),
            "ask1": float(field_list[21]),
            "ask2_vol": int(field_list[22]),
            "ask2": float(field_list[23]),
            "ask3_vol": int(field_list[14]),
            "ask3": float(field_list[25]),
            "ask4_vol": int(field_list[16]),
            "ask4": float(field_list[27]),
            "ask5_vol": int(field_list[28]),
            "ask5": float(field_list[29]),
        }
    except IndexError:
        return None
    return res


def exchange_prefix(code: str) -> str:
    return "sh" if code.startswith("6") else "sz"


def format_stock_code(code: str) -> str:
    if code[0].isdigit():
        return exchange_prefix(code) + code
    return code


def format_stock_codes(codes: list) -> list:
    if codes[0][0].isdigit():
        return [exchange_prefix(code) + code for code in codes]
    return codes



import os
import re

import setuptools


def get_version():
    with open(os.path.join("fastquotes", "__init__.py")) as f:
        content = f.read()
        return re.search('__version__ = "(.*?)"', content).group(1)


def strip_comments(comments):
    return comments.split("#", 1)[0].strip()


def reqs(*f):
    return list(
        filter(
            None,
            [
                strip_comments(s)
                for s in open(os.path.join(os.getcwd(), *f)).readlines()
            ],
        )
    )


setuptools.setup(
    name="fastquotes",
    version=get_version(),
    author="nocilantro",
    url="https://github.com/YangzhenZhao/fastquotes",
    install_requires=reqs("requirements.txt"),
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points="""
        [console_scripts]
        fquotes=fastquotes.cli:cli
    """,
)



import click

from .fund import fund_intro_dict
from .fund.daily import fund_real_time_dict


@click.command()
@click.option("-c", "--codes", multiple=True, help="Set fund code to get profit.")
@click.option("-i", "--intro", multiple=True, help="Set fund code to get intro.")
def cli(codes, intro):
    if codes:
        fund_dic = fund_real_time_dict()
        for code in codes:
            item = fund_dic[code]
            profit = float(item["单位净值"]) / float(item["上个交易日单位净值"]) - 1
            print(code, item["update_date"], profit)
    if intro:
        dic = fund_intro_dict()
        for code in intro:
            print(dic[code])
